@@ -17,7 +17,7 @@ import shlex
 
 from flask import json
 
-import flask, redis, bs4
+import flask, redis, bs4, werkzeug
 
 import potluck.time_utils, potluck.html_tools, potluck.render
 
@@ -111,7 +111,7 @@ def logs_folder(course, semester, username):
     """
     The logs folder for a class/semester/user.
     """
-    return flask.safe_join(
+    return werkzeug.utils.safe_join(
         evaluation_directory(course, semester),
         "logs",
         username
@@ -122,7 +122,7 @@ def reports_folder(course, semester, username):
     """
     The reports folder for a class/semester/user.
     """
-    return flask.safe_join(
+    return werkzeug.utils.safe_join(
         evaluation_directory(course, semester),
         "reports",
         username
@@ -155,7 +155,17 @@ def task_info_file(course, semester):
     """
     return os.path.join(
         evaluation_directory(course, semester),
-        _CONFIG["TASK_INFO_FILE"]
+        _CONFIG.get("TASK_INFO_FILE", "tasks.json")
+    )
+
+
+def concepts_file(course, semester):
+    """
+    The concepts file for a class/semester.
+    """
+    return os.path.join(
+        evaluation_directory(course, semester),
+        _CONFIG.get("CONCEPTS_FILE", "concepts.json")
     )
 
 
@@ -253,6 +263,30 @@ def evaluation_key(course, semester, username, pset, phase, task):
     )
 
 
+def exercise_key(course, semester, username, exercise):
+    """
+    The Redis key for the outcomes-list-history for a particular
+    exercise, submitted by a user for a particular course/semester.
+    """
+    return redis_key(
+        ':'.join(
+            [course, semester, "outcomes", username, exercise]
+        )
+    )
+
+
+def best_exercises_key(course, semester, username):
+    """
+    The Redis key for the per-exercise-best-outcomes-list data for a
+    particular user in a course/semester.
+    """
+    return redis_key(
+        ':'.join(
+            [course, semester, "exercises", username]
+        )
+    )
+
+
 #----------------#
 # Roster loading #
 #----------------#
@@ -316,6 +350,15 @@ def load_roster_from_stream(iterable_of_strings):
     """
     Implements the roster-loading logic given an iterable of strings,
     like an open file or a list of strings. See `AsRoster`.
+
+    Each entry in the dictionary it returns has the following keys:
+
+    - 'username': The student's username
+    - 'fullname': The student's full name (as it appears in the roster
+        file, so not always what they want to go by).
+    - 'sortname': A version of their name with 'last' name first that is
+        used to sort students.
+    - 'course_section': Which section the student is in.
     """
     reader = csv.reader(iterable_of_strings)
 
@@ -432,9 +475,48 @@ def get_task_info(course, semester):
             # Graft static task info into pset task entry
             task.update(result["tasks"][task["id"]])
 
+    # Augment exercise info
+    exfmtmap = _CONFIG.get("EXERCISE_URLS", {}).get(course, {})
+    for egroup in result["exercises"]:
+        if 'url' not in egroup:
+            for prefix in exfmtmap:
+                if egroup["group"].startswith(prefix):
+                    egroup["url"] = exfmtmap[prefix].format(
+                        semester=semester,
+                        group=egroup["group"]
+                    )
+
+        # No prefix matches perhaps
+        if 'url' not in egroup:
+            egroup["url"] = '#'
+
     return result
 
 
+def get_concepts(course, semester):
+    """
+    Loads concepts from the JSON file (or returns a cached version if the
+    file hasn't been modified since we last loaded it). Needs the course
+    and semester to load info for.
+
+    Returns None if the file doesn't exist or can't be parsed.
+    """
+    filename = concepts_file(course, semester)
+    try:
+        return load_or_get_cached(
+            filename,
+            assume_fresh=_CONFIG.get("ASSUME_FRESH", 1)
+        )
+    except Exception:
+        flask.flash("Failed to read concepts file!")
+        tb = potluck.html_tools.string_traceback()
+        print(
+            "Failed to read concepts file:\n" + tb,
+            file=sys.stderr
+        )
+        return None
+
+
 def get_admin_info(course, semester):
     """
     Reads the admin info file to get information about which users are
@@ -461,7 +543,8 @@ def get_admin_info(course, semester):
 def get_roster(course, semester):
     """
     Loads and returns the roster file. Returns None if the file is
-    missing.
+    missing. Returns a dictionary where usernames are keys and values are
+    student info (see `AsRoster`).
     """
     return load_or_get_cached(
         roster_file(course, semester),
@@ -827,14 +910,14 @@ def put_inflight(course, semester, username, phase, psid, taskid):
             log_folder = logs_folder(course, semester, username)
             ensure_directory(log_folder)
             logfile = unused_filename(
-                flask.safe_join(log_folder, istring + ".log")
+                werkzeug.utils.safe_join(log_folder, istring + ".log")
             )
 
             # Select an unused report filename
             report_folder = reports_folder(course, semester, username)
             ensure_directory(report_folder)
             reportfile = unused_filename(
-                flask.safe_join(report_folder, istring + ".json")
+                werkzeug.utils.safe_join(report_folder, istring + ".json")
             )
 
             # Gather the info into a tuple
@@ -1214,6 +1297,281 @@ def set_evaluation(
     return True
 
 
+def fetch_outcomes(course, semester, username, exercise):
+    """
+    Fetches the outcomes-list information for the given
+    user/course/semester/exercise. The result will be a list of
+    dictionaries each representing a single submission, in chronological
+    order. Each will have the following keys:
+
+    - "submitted_at" - A time string (see
+        `potluck.time_utils.timestring`) indicating when the list of
+        outcomes was submitted.
+    - "authors" - A list of usernames for participating authors.
+    - "outcomes" - A list of 3-tuple outcomes, which contain a boolean
+        for success/failure followed by tag and message strings (see
+        `optimism.listOutcomesInSuite`)
+    - "code": A list of filename, code pairs with any code blocks
+        submitted along with the outcomes.
+    - "status": A status string indicating the overall exercise status,
+        determined by callers to `save_outcomes`.
+    - "credit": A number indicating how much credit (0-1) was earned for
+        this outcome. Note that the best-credit outcome is retained in
+        the best-outcomes dictionary, with ties broken towards
+        more-recent submissions.
+
+    Returns None instead of a list if there is no information for that
+    user/exercise yet, or if an error is encountered while trying to
+    access that information.
+    """
+    # Redis key to use
+    key = exercise_key(course, semester, username, exercise)
+
+    try:
+        exists = _REDIS.exists(key)
+    except Exception:
+        flask.flash("Error checking for outcomes info.")
+        tb = potluck.html_tools.string_traceback()
+        print(
+            "Failed to check for outcomes info at '{}':\n{}".format(
+                key,
+                tb
+            ),
+            file=sys.stderr
+        )
+        return None
+
+    # Return None without making a fuss if the key just doesn't exist
+    if not exists:
+        return None
+
+    try:
+        responseJSON = _REDIS.get(key)
+        info = json.loads(responseJSON)
+    except Exception:
+        flask.flash("Error fetching or decoding outcomes info.")
+        tb = potluck.html_tools.string_traceback()
+        print(
+            "Failed to fetch outcomes info at '{}':\n{}".format(
+                key,
+                tb
+            ),
+            file=sys.stderr
+        )
+        return None
+
+    return info
+
+
+def fetch_best_outcomes(course, semester, username):
+    """
+    Fetches the best outcomes-list information for all exercises for the
+    given user/course/semester. The result will be a dictionary which
+    maps exercise IDs to dictionaries each with the following keys:
+
+    - "submitted_at" - A time string (see
+        `potluck.time_utils.timestring`) indicating when the list of
+        outcomes was submitted.
+    - "authors" - A list of usernames for participating authors.
+    - "outcomes" - A list of 3-tuple outcomes, which contain a boolean
+        for success/failure followed by tag and message strings (see
+        `optimism.listOutcomesInSuite`)
+    - "code" - A list of filename/code-string pairs indicating any code
+        blocks attached to the submission.
+    - "status" - A status string describing the exercise status based on
+        the outcomes which passed/failed.
+    - "credit" - A number indicating how much credit this exercise is
+        worth (higher = better) or possibly `None` if there is some issue
+        with the submission (like wrong # of outcomes).
+
+    Only the IDs of exercises the user has submitted at least once will
+    appear in the base dictionary.
+
+    Returns None if an error is encountered, but otherwise returns at
+    least an empty dictionary, including when the specified user hasn't
+    submitted anything yet.
+    """
+    # Redis key to use
+    key = best_exercises_key(course, semester, username)
+
+    try:
+        exists = _REDIS.exists(key)
+    except Exception:
+        flask.flash("Error checking for best-outcomes info.")
+        tb = potluck.html_tools.string_traceback()
+        print(
+            "Failed to check for best-outcomes info at '{}':\n{}".format(
+                key,
+                tb
+            ),
+            file=sys.stderr
+        )
+        return None
+
+    # Return an empty dictionary without making a fuss if the key just
+    # doesn't exist
+    if not exists:
+        return {}
+
+    try:
+        allJSON = _REDIS.hgetall(key)
+        result = {key: json.loads(allJSON[key]) for key in allJSON}
+    except Exception:
+        flask.flash("Error fetching or decoding best-outcomes info.")
+        tb = potluck.html_tools.string_traceback()
+        print(
+            "Failed to fetch best-outcomes info at '{}':\n{}".format(
+                key,
+                tb
+            ),
+            file=sys.stderr
+        )
+        return None
+
+    return result
+
+
+def save_outcomes(
+    course,
+    semester,
+    username,
+    exercise,
+    authors,
+    outcomes,
+    codeBlocks,
+    status,
+    credit
+):
+    """
+    Saves a list of outcomes for a specific exercise submitted by a
+    particular user who is taking a course in a certain semester. The
+    outcomes list should be a list of 3-tuples each consisting of a
+    boolean, a tag string, and a message string (e.g., the return value
+    from `optimism.listOutcomesInSuite`). The authors value should be a
+    list of username strings listing all authors who contributed. The
+    `codeBlocks` value should be a list of pairs, each of which has a
+    filename string and a code string (the filename could also elsehow
+    identify the source code was derived from). The `status` value should
+    be a string describing the status of the submission, while the credit
+    value should be a number where a higher number indicates a better
+    submission (to be stored preferentially as the "best" submission so
+    far if it's better than the old one).
+
+    The list of outcomes and associated code blocks is added to the
+    record of all such lists submitted for that exercise by that user, as
+    a dictionary with the following slots:
+
+    - "submitted_at": the current time, as a string (see
+        `potluck.time_utils.timestring`).
+    - "authors": The list of authors.
+    - "outcomes": The list of outcomes.
+    - "code": The list of filename/code-string pairs.
+    - "status": The status string.
+    - "credit": The credit number.
+
+    It will also replace any old outcomes list which is listed in the
+    best-outcomes-per-exercise info for the submitting user if it's
+    better than or the same as (in terms of credit) the best one
+    previously recorded there. If `credit` is set to None, it will count
+    as -1 for this purpose.
+
+    Returns True if it succeeds or False if it encounters some sort of
+    error.
+    """
+    # Redis key to use
+    key = exercise_key(course, semester, username, exercise)
+
+    # Generate a timestamp for the info
+    timestring = potluck.time_utils.timestring()
+
+    # Get old outcomes so we can add to them
+    recorded_outcomes = fetch_outcomes(
+        course,
+        semester,
+        username,
+        exercise
+    )
+    if recorded_outcomes is None:
+        recorded_outcomes = []
+
+    # Here's the info we store
+    info = {
+        "submitted_at": timestring,
+        "authors": authors,
+        "outcomes": outcomes,
+        "code": codeBlocks,
+        "status": status,
+        "credit": credit
+    }
+
+    recorded_outcomes.append(info)
+
+    new_encoded = json.dumps(recorded_outcomes)
+
+    try:
+        success = _REDIS.set(key, new_encoded)
+        if success is not True:
+            raise ValueError("Redis result indicated failure.")
+    except Exception:
+        flask.flash("Failed to write outcomes info!")
+        tb = potluck.html_tools.string_traceback()
+        print(
+            "Failed to write outcomes info at '{}':\n{}".format(
+                key,
+                tb
+            ),
+            file=sys.stderr
+        )
+        return False
+
+    # Also store info in best-outcomes map if it is good
+    best_key = best_exercises_key(course, semester, username)
+    # TODO: We should pipeline this with the other write to try to more
+    # completely avoid situations where the first succeeds and the second
+    # fails!
+    try:
+        old_json = _REDIS.hget(best_key, exercise)
+        if old_json is None:
+            old_credit = -1
+        else:
+            old_best = json.loads(old_json)
+            old_credit = old_best.get("credit", -1)
+    except Exception:
+        flask.flash("Failed to read best-outcomes info!")
+        tb = potluck.html_tools.string_traceback()
+        print(
+            "Failed to read best-outcomes info at '{}':\n{}".format(
+                best_key,
+                tb
+            ),
+            file=sys.stderr
+        )
+        return False
+
+    # Missing credit value counts as -1 credit.
+    if credit is None:
+        credit = -1
+
+    # Only record a new best-outcome-for-this-exercise if the current
+    # credit value is >= the old one.
+    if credit >= old_credit:
+        try:
+            _REDIS.hset(best_key, exercise, json.dumps(info))
+        except Exception:
+            flask.flash("Failed to write best-outcomes info!")
+            tb = potluck.html_tools.string_traceback()
+            print(
+                "Failed to write best-outcomes info at '{}':\n{}".format(
+                    best_key,
+                    tb
+                ),
+                file=sys.stderr
+            )
+            return False
+
+    return True
+
+
 def default_feedback_summary():
     """
     Returns a default summary object. The summary is a pared-down version
@@ -1574,7 +1932,8 @@ class AsStudentInfo(View):
 class AsRoster(View):
     """
     Encoding and decoding for CSV rosters, which are cached. The roster
-    structure is a dictionary mapping usernames to student info.
+    structure is a dictionary mapping usernames to student info (see
+    `load_roster_from_stream`).
     """
     @staticmethod
     def encode(obj):
@@ -1694,8 +2053,15 @@ Cache of objects returned by view functions on cache keys.
 
 _FRESH_AT = {}
 """
-Mapping from cache keys to the most recent time at which they were found
-to be fresh.
+Mapping from cache keys to tuples of seconds-since-epoch floats
+representing the most recent time-span during which the file was found to
+be unchanged. The first element of each tuple represents the earliest time
+at which a freshness check succeeded with no subsequent failed checks,
+and the second element represents the most recent time another successful
+check was performed. Whenever a check is performed and succeeds, the
+second element is updated, and whenever a check is performed and fails,
+the value is set to None. When the value is None and a check succeeds,
+both elements of the tuple are set to the current time.
 """
 
 
@@ -1725,11 +2091,16 @@ def build_file_freshness_checker(
         """
         global _FRESH_AT
         now = time.time()
-        if (
-            cache_key in _FRESH_AT
-        and now - _FRESH_AT[cache_key] < assume_fresh
-        ):
-            return True
+        cached_fresh = _FRESH_AT.get(cache_key)
+        if cached_fresh is not None:
+            cached_start, cached_end = cached_fresh
+            if (
+                ts is not None
+            and ts >= cached_start
+                # Note we don't check ts <= cached_end here
+            and (now - cached_end) < _CONFIG.get("ASSUME_FRESH", 1)
+            ):
+                return False  # we assume it has NOT changed
 
         filename = cache_key_filename(cache_key)
         try:
@@ -1744,7 +2115,37 @@ def build_file_freshness_checker(
         # timestamp, or if the timestamp is None
         result = ts is None or mtime >= ts
         if result:
-            _FRESH_AT[cache_key] = now
+            # If we find that a specific time was checked, and that time
+            # was after the beginning of the current cached fresh period,
+            # we erase the cached fresh period, since result being true
+            # means the file HAS changed.
+            if (
+                ts is not None
+            and (
+                    cached_fresh is not None
+                and ts > cached_start
+                )
+            ):
+                _FRESH_AT[cache_key] = None
+        else:
+            # If the file WASN'T changed, we extend the cache freshness
+            # (In this branch ts is NOT None and it's strictly greater
+            # than mtime)
+            if cached_fresh is not None:
+                # If an earlier-time-point was checked and the check
+                # succeeded, we can extend the fresh-time-span backwards
+                if ts < cached_start:
+                    cached_start = ts
+
+                # Likewise, we might be able to extend it forwards
+                if ts > cached_end:
+                    cached_end = ts
+
+                # Update the time-span
+                _FRESH_AT[cache_key] = (cached_start, cached_end)
+            else:
+                # If we didn't have cached freshness, initialize it
+                _FRESH_AT[cache_key] = (ts, ts)
 
         return result
 
